import asyncio
import importlib
import logging
import os
import hashlib
import secrets

from aiohttp import web

import config
import http_client
import paths
import web_interface.page_former
import user_login
import version


PATHS_WITH_FILE_UPLOAD = {'accept_update_from_file' : paths.UPDATE_FILE_PATH,\
                          'upload_plugin' :  paths.PLUGIN_INSTALL_FILE_PATH,\
                          'upload_to_storage' : None}
EXECUTOR_TIMEOUT = 5.8 # 0.2s less than WebInterface.SHUTDOWN_TIMEOUT
UPLOAD_CHUNK_SIZE = 1024*256 # greatly affects upload performance
AUTH_ERROR_SLEEP = 3
SHOW_TOKENS_IN_API = False


def check_auth(request):
    try:
        host = request.host.split(':')[0]
    except:
        host = "127.0.0.1"
    if request.remote != host:
        if request.secure:
            cookie = request.cookies.get(web_interface.page_former.PageFormer.COOKIE_NAME_HTTPS)
        else:
            cookie = request.cookies.get(web_interface.page_former.PageFormer.COOKIE_NAME)
        if not cookie or cookie not in request.app['cookie_tokens']:
            return "Error: uploads allowed only with valid token cookie"


def find_printer_interface(app, printer_id):
    for pi in app.printer_interfaces:
        if not printer_id or pi.id_string == printer_id:
            return pi


def printer_report(pi):
    report = {}
    report['report'] = pi.status_report()
    report['errors'] = pi.get_errors_to_send()
    http_client.remove_repeted_errors(report)
    reg_code = pi.registration_code
    if reg_code:
        report['reg_code'] = reg_code
    if SHOW_TOKENS_IN_API:
        report['token'] = pi.printer_token
    return report


def get_static_values(app):
    values = {'version' : version.full_version_string()}
    values['cloud_url'] = config.get_settings()['URL'].strip("cli-")
    values['style_name'] = config.get_settings().get("web_interface", {}).get("style", "style")
    values['show_your_account_button'] = not app.offline_mode 
    return values


def full_report(app):
    response_dict = {'printers': {}}
    for pi in app.printer_interfaces:
        response_dict['printers'][pi.id_string] = printer_report(pi)
        if getattr(app, "updater"):
            response_dict['update_available'] = app.updater.update_available
            response_dict['update_downloaded'] = app.updater.update_downloaded
    return response_dict


def printer_command(path_list, args, app):
    error = ''
    status = web.HTTPOk.status_code
    if len(path_list) > 1:
        printer_id = path_list[1]
    else:
        printer_id = args.get('printer_id')
    if not printer_id:
        try:
            pi = app.printer_interfaces[0]
        except IndexError:
            pi = None
    else:
        pi = find_printer_interface(app, printer_id)
        if not pi:
            error == 'No printer with id ' + str(printer_id)
            status = web.HTTPClientError.status_code
    if pi:
        if 'command' in args or path_list[2] == 'command':
            if len(path_list) > 2:
                command = path_list[3]
            else:
                command = args.get('command')
            if command:
                command_args = {'command': command, 'number': -1}
                payload = {}
                payload.update(args)
                if 'command' in payload:
                    del payload['command']
                if 'printer_id' in payload:
                    del payload['printer_id']
                if payload:
                    command_args['payload'] = payload
                ack = pi.execute_server_command(command_args)
                if not ack['result']:
                    body = {'error': {'message': 'Command failure',
                                      'code': web.HTTPServerError.status_code,
                                      'ack' : ack}}
                else:
                    body = ack
            else:
                status = web.HTTPClientError.status_code
                error = 'No command'
        elif len(path_list) < 3 or path_list[2] == 'report' or 'report' in args or path_list[2] == '':
            body = printer_report(pi)
            if not body:
                status = web.HTTPClientError.status_code
                error = 'The printer has disconnected'
        else:
            status = web.HTTPClientError.status_code
            error = 'No such command'
    return error, status


async def generic(request):
    logger = logging.getLogger('app.generic_web_handler')
    pf_args = []
    if request.method == 'GET':
        logger.debug("GET %s" % request.path)
        method_name = 'do_GET'
    elif request.method == 'POST':
        logger.debug("POST %s" % request.path)
        method_name = 'do_POST'
        post_data = await request.post()
        pf_args.append(post_data)
    else:
        raise web.HTTPForbidden()
    return await form_page_and_respond(request, method_name, pf_args)


async def reload_page_former(request):
    if request.remote == request.host.split(':')[0]:
        logging.getLogger('app.reload_page_former').info("Reloading page forment")
        importlib.reload(web_interface.page_former)
        raise web.HTTPNoContent


async def multipart_upload_handler(request):
    logger = logging.getLogger('app.multipart_upload_handler')
    error = check_auth(request)
    file_path = PATHS_WITH_FILE_UPLOAD.get(request.path.strip("/"))
    logger.info("Upload path: %s to %s" % (request.path, file_path)) 
    if not error:
        try:
            reader = await request.multipart()
            fields_left = 128 # max number of fields
            while True:
                if request.app['3dp_app'].stop_flag:
                    raise web.HTTPFound("/")
                field = await reader.next()
                if field and field.name == 'file':
                    break
                if fields_left:
                    fields_left -= 1
                else:
                    raise RuntimeError('No upload got not file field')
        except Exception as e:
            error = "Error while searching for upload field: " + str(e)
            logger.error(error)
        else:
            if not file_path:
                filename = field.filename
                if filename:
                    file_path = os.path.join(paths.STORAGE_FOLDER, filename)
                else:
                    error = "Please select a file"
    if not error:
        size = 0
        try:
            with open(file_path, 'wb') as f:
                while True:
                    if request.app['3dp_app'].stop_flag:
                        break
                    chunk = await field.read_chunk(UPLOAD_CHUNK_SIZE) 
                    if not chunk:
                        break
                    size += len(chunk)
                    file_write_future = asyncio.get_event_loop().run_in_executor(
                            request.app['threads_pool'],
                            lambda f, chunk: f.write(chunk),
                            f,
                            chunk)
                    await asyncio.wait_for(file_write_future, EXECUTOR_TIMEOUT)
            if request.app['3dp_app'].stop_flag:
                os.remove(file_path)
                raise web.HTTPFound("/")
        except (OSError, IOError) as e:
            error = "Error saving upload: " + str(e)
            logger.error(error)
        else:
            logger.info('Received upload size: %dB' % size)
    pf_args = [{"multipart_error" : error}]
    return await form_page_and_respond(request, 'do_POST', pf_args)


async def form_page_and_respond(request, method_name, pf_args=[]):
    logger = logging.getLogger('app.generic_web_handler')
    pf = web_interface.page_former.PageFormer(request.app['3dp_app'],
                     request.app['3dp_content'],
                     request.host,
                     request.method,
                     request.raw_path,
                     request.secure,
                     request.remote,
                     request.cookies,
                     request.app['cookie_tokens'],
                     request.app['wizard_stage'],
                     request.app['wizard_enabled'])
    try:
        executor_future = asyncio.get_event_loop().run_in_executor(request.app['threads_pool'], getattr(pf, method_name), *pf_args)
        await asyncio.wait_for(executor_future, EXECUTOR_TIMEOUT)
    except asyncio.TimeoutError:
        logger.warning(f'Executor timeout when handling {request.raw_path}')
        pf.status = 504
    except Exception as e:
        logger.error(e) # check why some error are not logged by log.log_exception
    else:
        if pf.wizard_former.current_stage != request.app['wizard_stage']:
            logger.info("Wizard stage: " + str(pf.wizard_former.current_stage))
            request.app['wizard_stage'] = pf.wizard_former.current_stage
        request.app['wizard_enabled'] = pf.wizard_enabled
        if pf.new_cookie_token:
            if pf.new_cookie_token not in request.app['cookie_tokens']:
                request.app['cookie_tokens'].append(pf.new_cookie_token)
                request.app.save_cookie_tokens()
    return web.Response(status=pf.status, headers=pf.headers, body=pf.body)


def process_login(args, app):
    login = args.get('login', '').replace(' ', '+') # a hack to fix + in email.
    password = args.get('password', '')
    password_hash = args.get('password_hash')
    print("!!!args", args)
    if not password_hash:
        password_hash = hashlib.sha256(password.encode('utf-8')).hexdigest()
    auth_token = args.get('auth_token')
    if not login and not auth_token:
        return "No login"
    correct_login, correct_password_hash = user_login.UserLogin.load_login()
    if correct_login:
        if login != correct_login or password_hash != correct_password_hash:
            return "Incorrect credentials"
    elif not app.user_login.got_network_connection:
        return "No network to check login credentials on the server"
    else:
        error = app.user_login.login_as_user(login, password_hash, auth_token)
        if error:
            return error[1]


async def api2_handler(request):
    body = {}
    path_list = request.path.split('/')[2:]
    status = web.HTTPOk.status_code
    args = {}
    args.update(request.query)
    app = request.app['3dp_app']
    if request.method == 'POST':
        if request.content_type == 'application/json':
            args.update(await request.json())
        else:
            args.update(await request.post())
    print("!", path_list, args)
    if path_list and path_list[0] == 'login':
        error = process_login(args, request.app['3dp_app'])
        if error:
            status = web.HTTPNonAuthoritativeInformation.status_code
            await asyncio.sleep(AUTH_ERROR_SLEEP)
        else:
            new_cookie_token = secrets.token_hex(web_interface.page_former.PageFormer.COOKIE_TOKEN_LENGHT)
            request.app['cookie_tokens'].append(new_cookie_token)
            cookie_header_value = f"token:{new_cookie_token}; HttpOnly"
            if request.secure:
                cookie_header_value += '; Secure'
            body = {'auth_token': new_cookie_token}
            return web.json_response(status=status,
                                 data=body,
                                 headers={'Set-cookie': cookie_header_value})
    else:
        error = check_auth(request)
        provided_auth_token = args.get('auth_token')
        if not error or (provided_auth_token and provided_auth_token in request.app['cookie_tokens']):
            if 'full_report' in path_list:
                body = full_report(app)
            elif 'static_values' in path_list:
                body = get_static_values(app)
            elif 'settings' in path_list:
                if request.method == "GET":
                    body = config.get_settings()
                elif request.method == "POST":
                    settings = config.get_settings()
                    try:
                        config.merge_dictionaries(settings, args)
                        config.Config.instance().save_settings(settings)
                    except Exception as e:
                        error = "Error on saving settings"
                        status = web.HTTPServerError
            elif 'printer_id' in args or 'printer_id' in path_list:
                error, status = printer_command(path_list, args, app)
            elif 'logout' in path_list:
                app.user_login.logout()
                app.restart_flag = True
                app.stop_flag = True
        else:
            status = web.HTTPUnauthorized.status_code
    if error:
        body = {'error': {'message': error, 'code': status}}
    return web.json_response(status=status, data=body)
